INTERFACE [arm && fpu]:

EXTENSION class Fpu
{
public:
  
  struct Exception_state_user
    {
      Mword fpexc;
      Mword fpinst;
      Mword fpinst2;
    };

private:

  enum
    {
      FPEXC_EN  = 1 << 30,
      FPEXC_EX  = 1 << 31,
    };

  struct Fpu_regs
    {
      Mword fpexc, fpscr;
      Mword state[32 * 4]; // 4*32 bytes for each FP-reg
    };

  static Mword fpsid_rev(Mword v)          { return v & 0xf; }
  static Mword fpsid_variant(Mword v)      { return (v >> 4) & 0xf; }
  static Mword fpsid_part_number(Mword v)  { return (v >> 8) & 0xff; }
  static Mword fpsid_arch_version(Mword v) { return (v >> 16) & 0xf; }
  static Mword fpsid_precision(Mword v)    { return (v >> 20) & 1; }
  static Mword fpsid_format(Mword v)       { return (v >> 21) & 3; }
  static Mword fpsid_hw_sw(Mword v)        { return (v >> 23) & 1; }
  static Mword fpsid_implementer(Mword v)  { return v >> 24; }
};

// ------------------------------------------------------------------------
INTERFACE [arm && !fpu]:

EXTENSION class Fpu
{
public:
  struct Exception_state_user
    {
    };
};

// ------------------------------------------------------------------------
INTERFACE [arm]:

EXTENSION class Fpu
{
public:
  enum Fpu_exception_returns
    {
      Fpu_except_none,
      Fpu_except_emulated,
      Fpu_except_fault,
    };
};

// ------------------------------------------------------------------------
IMPLEMENTATION [arm && !fpu]:

#include "trap_state.h"

PUBLIC static inline NEEDS["trap_state.h"]
void
Fpu::save_user_exception_state(Trap_state *, Exception_state_user *)
{}

PUBLIC static inline NEEDS["trap_state.h"]
Fpu::Fpu_exception_returns
Fpu::handle_fpu_trap(Trap_state *)
{ return Fpu::Fpu_except_none; }

PUBLIC static inline
bool
Fpu::is_enabled()
{ return false; }

PUBLIC static inline
bool
Fpu::exc_pending()
{ return false; }

// ------------------------------------------------------------------------
IMPLEMENTATION [arm && fpu]:

#include <cassert>
#include <cstdio>
#include <cstring>

#include "fpu_state.h"
#include "mem.h"
#include "processor.h"
#include "static_assert.h"
#include "trap_state.h"

PRIVATE static inline
Mword
Fpu::fpsid()
{
  Mword v;
  //asm volatile("fmrx %0, fpsid" : "=r" (v));
  asm volatile("mrc p10, 7, %0, cr0, cr0, 0" : "=r" (v));
  return v;
}

PRIVATE static inline
void
Fpu::fpexc(Mword v)
{
  //asm volatile("fmxr fpexc, %0" : : "r" (v));
  asm volatile("mcr p10, 7, %0, cr8, cr0, 0" : : "r" (v));
}

PUBLIC static inline
Mword
Fpu::fpexc()
{
  Mword v;
  //asm volatile("fmrx %0, fpexc" : "=r" (v));
  asm volatile("mrc p10, 7, %0, cr8, cr0, 0" : "=r" (v));
  return v;
}

PUBLIC static inline
Mword
Fpu::fpinst()
{
  Mword i;
  asm volatile("mcr p10, 7, %0, cr9,  cr0, 0" : "=r" (i));
  return i;
}

PUBLIC static inline
Mword
Fpu::fpinst2()
{
  Mword i;
  asm volatile("mcr p10, 7, %0, cr10,  cr0, 0" : "=r" (i));
  return i;
}

PUBLIC static inline
bool
Fpu::exc_pending()
{
  return fpexc() & FPEXC_EX;
}

IMPLEMENT
void
Fpu::enable()
{
  fpexc((fpexc() | FPEXC_EN) & ~FPEXC_EX);
}

IMPLEMENT
void
Fpu::disable()
{
  fpexc(fpexc() & ~FPEXC_EN);
}



IMPLEMENT
void
Fpu::init(unsigned cpu)
{
  // XXX MPCORE: enable cp11, cp10
  asm volatile(" mcr  p15, 0, %0, c1, c0, 2   \n" : : "r"(0x00f00000));
  Mem::dsb();

  Mword s = fpsid();
  printf("FPU%d: Arch: %s(%lx), Part: %s(%lx), r: %lx, v: %lx, i: %lx, t: %s, p: %s\n",
         cpu, fpsid_arch_version(s) == 1
           ? "VFPv2"
           : (fpsid_arch_version(s) == 3 ? "VFPv3" : "Unkn"),
         fpsid_arch_version(s),
         fpsid_part_number(s) == 0x20
           ? "VFP11"
           : (fpsid_part_number(s) == 0x30 ?  "VFPv3" : "Unkn"),
         fpsid_part_number(s),
         fpsid_rev(s), fpsid_variant(s), fpsid_implementer(s),
         fpsid_hw_sw(s) ? "soft" : "hard",
         fpsid_precision(s) ? "sngl" : "dbl/sngl");

  disable();

  set_owner(cpu, 0);
}

IMPLEMENT inline NEEDS ["fpu_state.h", "mem.h", "static_assert.h", <cstring>]
void
Fpu::init_state (Fpu_state *s)
{
  Fpu_regs *fpu_regs = reinterpret_cast<Fpu_regs *>(s->state_buffer());
  static_assert(!(sizeof (*fpu_regs) % sizeof(Mword)),
                "Non-mword size of Fpu_regs");
  Mem::memset_mwords(fpu_regs, 0, sizeof (*fpu_regs) / sizeof(Mword));
}

IMPLEMENT
void
Fpu::save_state(Fpu_state *s)
{
  assert(s->state_buffer());
  Fpu_regs *fpu_regs = reinterpret_cast<Fpu_regs *>(s->state_buffer());

  asm volatile ("stc p11, cr0, [%0], #32*4     \n"
                : : "r" (fpu_regs->state));
  asm volatile ("mrc p10, 7, %0, cr8,  cr0, 0  \n"
                "mrc p10, 7, %1, cr1,  cr0, 0  \n"
                : "=r" (fpu_regs->fpexc),
                  "=r" (fpu_regs->fpscr));
}

IMPLEMENT
void
Fpu::restore_state (Fpu_state *s)
{
  assert (s->state_buffer());
  Fpu_regs *fpu_regs = reinterpret_cast<Fpu_regs *>(s->state_buffer());

  asm volatile ("ldc p11, cr0, [%0], #32*4     \n"
                : : "r" (fpu_regs->state));
  asm volatile ("mcr p10, 7, %0, cr8,  cr0, 0  \n"
                "mcr p10, 7, %1, cr1,  cr0, 0  \n"
                :
                : "r" (fpu_regs->fpexc | FPEXC_EN),
                  "r" (fpu_regs->fpscr));

#if 0
  asm volatile("mcr p10, 7, %2, cr9,  cr0, 0  \n"
               "mcr p10, 7, %3, cr10, cr0, 0  \n"
               :
               : "r" (fpu_regs->fpinst),
                 "r" (fpu_regs->fpinst2));
#endif
}

IMPLEMENT inline
unsigned
Fpu::state_size()
{ return sizeof (Fpu_regs); }

IMPLEMENT inline
unsigned
Fpu::state_align()
{ return 4; }

PUBLIC static
bool
Fpu::is_enabled()
{
  return fpexc() & FPEXC_EN;
}

PRIVATE static inline NEEDS[Fpu::fpexc, "processor.h", "trap_state.h"]
bool
Fpu::is_fpu_trap(Trap_state *ts, Mword *opcode)
{
  if (ts->psr & Proc::Status_thumb)
    {
      Unsigned16 *insn = reinterpret_cast<Unsigned16*>(ts->ip()) - 1;

      if ((*insn & 0xe000) == 0xe000 && (*insn & 0x1800))
	*opcode = (Mword(*insn) << 16) | *(insn + 1);
      else
	return false;
    }
  else
    *opcode = *(reinterpret_cast<Mword*>(ts->ip()) - 1);

  if ((*opcode & 0xec000000) == 0xec000000
      && (((*opcode & 0x00000f00) == 0x00000a00)
          || ((*opcode & 0x00000f00) == 0x00000b00)))
    return true;

  return false;
}

PUBLIC static
Fpu::Fpu_exception_returns
Fpu::handle_fpu_trap(Trap_state *ts)
{
  Mword opcode;
  if (Fpu::is_fpu_trap(ts, &opcode))
    {
      if (Fpu::is_enabled())
	{
	  if ((opcode & 0x0fff0f90) == 0x0ef00a10)
	    {
	      unsigned reg = (opcode >> 12) & 0xf;
	      ts->r[reg] = Fpu::fpsid();
	      return Fpu_except_emulated;
	    }
	}
      return Fpu_except_fault;
    }

  return Fpu_except_none;
}

PUBLIC static inline NEEDS["trap_state.h"]
void
Fpu::save_user_exception_state(Trap_state *ts, Exception_state_user *esu)
{
  if ((ts->error_code & 0x01f00000) == 0x01100000)
    {
      esu->fpexc = Fpu::fpexc();
      if (ts->error_code == 0x03100000)
	{
	  esu->fpinst  = Fpu::fpinst();
	  esu->fpinst2 = Fpu::fpinst2();
	}
    }
}
