/**
 * \file
 */
/*
 * (c) 2008-2009 Technische Universität Dresden
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 *
 * As a special exception, you may use this file as part of a free software
 * library without restriction.  Specifically, if other files instantiate
 * templates or use macros or inline functions from this file, or you compile
 * this file and link it with other files to produce an executable, this
 * file does not by itself cause the resulting executable to be covered by
 * the GNU General Public License.  This exception does not however
 * invalidate any other reasons why the executable file might be covered by
 * the GNU General Public License.
 */
#pragma once

#include <l4/sys/types.h>
#include <l4/re/c/dataspace.h>

/**
 * \defgroup api_l4re_c_fb Frame-buffer API
 * \ingroup api_l4re_c
 */

/**
 * \brief Flags of information about the frame buffer.
 * \ingroup api_l4re_c_fb
 */
enum l4re_fb_info_flags_t
{
  F_l4re_fb_auto_refresh = 1 << 0, ///< 1 if framebuffer does auto-refreshing
  F_l4re_fb_has_pointer  = 1 << 1, ///< 1 if the framebuffer provider provides a point
};

/**
 * \brief Framebuffer information structure
 * \ingroup api_l4re_c_fb
 */
typedef struct
  {
    unsigned long base_offset;    ///< Offset to first pixel in data space
    unsigned long mem_total;      ///< Size of the frame buffer
    unsigned long map_size;       ///< Do not use, use fb_ds.size() instead
    unsigned x_res;               ///< Width of the frame buffer
    unsigned y_res;               ///< Height of the frame buffer
    unsigned bytes_per_scan_line; ///< Bytes per line
    unsigned flags;               ///< Flags of the framebuffer
    char bits_per_pixel;          ///< Bits per pixel
    char bytes_per_pixel;         ///< Bytes per pixel
    char red_size;                ///< Size of red color component in a pixel
    char red_shift;               ///< Shift of the red color component in a pixel
    char green_size;              ///< Size of green color component in a pixel
    char green_shift;             ///< Shift of the green color component in a pixel
    char blue_size;               ///< Size of blue color component in a pixel
    char blue_shift;              ///< Shift of the blue color component in a pixel
  } l4re_fb_info_t;

/**
 * \brief Framebuffer object type
 * \ingroup api_l4re_c_fb
 */
typedef l4_cap_idx_t l4re_fb_t;

EXTERN_C_BEGIN

/**
 * \brief Open a dataspace.
 * \ingroup api_l4re_c_fb
 *
 * \param  fb   Framebuffer object
 * \retval info Pointer to framebuffer information structure.
 *
 * \return 0 for success, <0 on error
 *         - -#L4_ENODEV
 *         - IPC errors
 */
L4_CV int
l4re_fb_info(l4re_fb_t fb, l4re_fb_info_t *info) L4_NOTHROW;

/**
 * \brief Get framebuffer memory dataspace.
 * \ingroup api_l4re_c_fb
 *
 * \param  fb   Framebuffer object
 * \param  ds   Capability index to put framebuffer dataspace into.
 *
 * \return 0 for success, <0 on error
 *         - -#L4_ENODEV
 *         - IPC errors
 */
L4_CV int
l4re_fb_mem(l4re_fb_t fb, l4re_ds_t ds) L4_NOTHROW;

/**
 * \brief Refresh frame buffer area.
 * \ingroup api_l4re_c_fb
 *
 * \param fb  Framebuffer object
 * \param x   X coordinate of area
 * \param y   Y coordinate of area
 * \param w   Width of area
 * \param h   Height of area
 *
 * \return 0 for success, <0 on error
 *         - -#L4_ENODEV
 *         - IPC errors
 */
L4_CV int
l4re_fb_refresh(l4re_fb_t fb, int x, int y, int w, int h) L4_NOTHROW;

EXTERN_C_END
