/*
 * (c) 2008-2009 Technische Universität Dresden
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU Lesser General Public License 2.1.
 * Please see the COPYING-LGPL-2.1 file for details.
 */
#include <l4/libloader/script>
#include <l4/re/util/name_space_svr>

#include <cstring>

namespace Ldr {

void
Parser::check_unqualified_name(Ldr::Token const &t)
{
  String::Index x = t.text().find("/");
  if (!t.text().eof(x))
    throw Parse_exception(_scan, t, "expected unqualified name, got:");
}


cxx::String
Parser::string_expr()
{
  if (cur().type() != Ldr::Token::String)
    throw Ldr::Parse_exception(_scan, cur(), "expected string, got:");

  cxx::String val = cur().text();
  shift();
  return val;
}

cxx::String
Parser::name_expr()
{
  if (cur().type() != Ldr::Token::Name)
    throw Ldr::Parse_exception(_scan, cur(), "expected name, got:");

  cxx::String val = cur().text();
  shift();
  return val;
}

void
Parser::param_list_start()
{
  if (cur().type() != Ldr::Token::L_paren)
    throw Parse_exception(_scan, last(), "excpected '(' after:");

  shift();
  return;
}

void 
Parser::param_list_end()
{
  if (cur().type() != Ldr::Token::R_paren)
    throw Parse_exception(_scan, last(), "excpected ')' after:");

  return;
}

Ldr::Token
Parser::param_next()
{

  switch (cur().type())
    {
    case Ldr::Token::Number:
    case Ldr::Token::String: 
    case Ldr::Token::Name:
	{
	  Ldr::Token v = cur();
	  shift();
	  if (cur().type() == Ldr::Token::Comma)
	    {
	      shift();
	      return v;
	    }
	  return v;
	}
    case Ldr::Token::R_paren:
      return Ldr::Token();
    default:
      throw Parse_exception(_scan, cur(), "expected parameter, got");
    }
}

Parser::Val
Parser::method_invocation(Val const &obj)
{
  shift();
  if (cur().type() != Ldr::Token::Name)
    throw Parse_exception(_scan, last(1), "expected method name after:");

  return invoke(obj, cur());
}


unsigned
Parser::parse_flags()
{
  shift();
  if (cur().type() != Ldr::Token::Name)
    throw Parse_exception(_scan, cur(), "unexpected token:");

  unsigned res = 0;
  cxx::String rights = cur().text();
  if (!rights.eof(rights.find("s")))
    res |= L4Re::Util::Names::Obj::F_strong;
  if (!rights.eof(rights.find("w")))
    res |= L4Re::Util::Names::Obj::F_rw;
  return res;
}

Parser::Val
Parser::capability_rvalue()
{
  if (cur().type() != Ldr::Token::Name)
    throw Parse_exception(_scan, last(), "excpected qualified name after");

  if (!_obj_mgr)
    throw Parse_exception(_scan, last(), "capability rvalues are not supported");

  Val rhs = _obj_mgr->find_object(cur().text(), scan()->ctxt());
  shift();

  if (cur().type() == Ldr::Token::Pointsat)
    {
      rhs = method_invocation(rhs);
      shift();
    }

  unsigned flags = 0;
  if (cur().type() == Ldr::Token::Comma)
    {
      flags = parse_flags();
      shift();
    }

  if (cur().type() != Ldr::Token::Semicolon)
    throw Parse_exception(_scan, last(), "expected ';' after:");

  rhs.add_flags(flags);
  return rhs;
}

void
Parser::check_assignment()
{
  shift();
  if (cur().type() != Ldr::Token::Assign)
    throw Parse_exception(_scan, cur(), "missing '=', found");
}

void
Parser::expression()
{
  if (cur().type() == Ldr::Token::Name
      || cur().type() == Ldr::Token::String)
    {
      check_unqualified_name(cur());
      entry();
    }
}

void
Parser::statement()
{
  switch (cur().type())
    {
    case Ldr::Token::Semicolon: /* empty statement */
      return;
    case Ldr::Token::Name:
    case Ldr::Token::String:
      expression();
      return;
    default:
      throw Parse_exception(_scan, cur(), "syntax error, unexpected token");
    }
}


static
bool operator != (cxx::String const &a, char const *b)
{
  if ((int)strlen(b) != a.len())
    return true;

  return strncmp(a.start(), b, a.len()) != 0;
}

static
bool operator == (cxx::String const &a, char const *b)
{
  if ((int)strlen(b) != a.len())
    return false;

  return strncmp(a.start(), b, a.len()) == 0;
}

static
int __idx(cxx::String const &s,
          Ldr::Script::Bit_desc const *desc)
{
  int i = 0;
  while (desc[i].name && s != desc[i].name)
    ++i;

  if (!desc[i].name)
    return -1;

  return i;
}


unsigned long
Parser::enum_list(Script::Bit_desc const *desc)
{
  unsigned long v = 0;
  while (1)
    {
      if (cur().type() != Ldr::Token::Name)
	throw Ldr::Parse_exception(_scan, cur(), "expected name, got");

      int i = __idx(cur().text(), desc);
      if (i < 0)
	throw Ldr::Parse_exception(_scan, cur(), "unexpected value:");

      v |= desc[i].bit;

      shift();
      if (cur().type() != Ldr::Token::Comma)
	break;

      shift();
    }
  return v;
}

unsigned long
Parser::number()
{
  if (cur().type() != Ldr::Token::Number)
    throw Ldr::Parse_exception(_scan, cur(), "expected number, found");

  long n = cur().number();
  shift();
  return n;
}

Parser::Val
Parser::invoke(Parser::Val const &obj, Token const &func)
{
  if (func.text() != "open")
    throw Parse_exception(scan(), func, "undefined method:");

  shift();
  param_list_start();
  enum { Max_params = 5 };
  cxx::String params[Max_params];
  int num_params = 0;
  while (1)
    {
      Ldr::Token t = param_next();
      if (t.type() == Ldr::Token::Eof)
	break;

      params[num_params] = t.text();
      ++num_params;
      if (num_params >= Max_params)
	break;
    }
  param_list_end();

  if (!_obj_mgr)
    throw Parse_exception(scan(), func, "method invocation not supported:");

  if (_no_invokes)
    return Parser::Val(L4::Cap<void>(L4_INVALID_CAP), 0);
  return _obj_mgr->invoke(obj, num_params, params, scan()->ctxt());
}


bool
Parser::Value_parser_set::parse(Parser *parser, Token const &tag)
{
  Value_parser *dp = _set;
  while (dp)
    {
      if (tag.text() == dp->tag)
	{
	  dp->parse(parser);
	  return true;
	}
      dp = dp->next;
    }
  return false;
}

}
